@@ -554,7 +554,7 @@ static posix_poll(apr_pollset_t *pollset,
     DBG(4, "entered\n");
 
     if (timeout > 0) {
-        timeout /= 1000;
+        timeout = (timeout + 999) / 1000;
     }
     rv = poll(priv->pollset, pollset->nelts, timeout);
     (*num) = rv;
@@ -698,6 +698,7 @@ static apr_status_t asio_pollset_poll(apr_pollset_t *pollset,
             tv.tv_nsec = apr_time_usec(timeout) * 1000;
         } else {
             tv.tv_sec = INT_MAX;  /* block until something is ready */
+            tv.tv_nsec = 0;
         }
 
         DBG2(6, "nothing on the ready ring "
@@ -768,7 +769,7 @@ static apr_status_t asio_pollset_poll(apr_pollset_t *pollset,
     return rv;
 }  /* end of asio_pollset_poll */
 
-static apr_pollset_provider_t impl = {
+static const apr_pollset_provider_t impl = {
     asio_pollset_create,
     asio_pollset_add,
     asio_pollset_remove,
@@ -777,6 +778,6 @@ static apr_pollset_provider_t impl = {
     "asio"
 };
 
-apr_pollset_provider_t *apr_pollset_provider_aio_msgq = &impl;
+const apr_pollset_provider_t *apr_pollset_provider_aio_msgq = &impl;
 
 #endif /* HAVE_AIO_MSGQ */
