@@ -27,6 +27,7 @@ const struct testlist {
     {testdir},
     {testdso},
     {testdup},
+    {testencode},
     {testenv},
     {testescape},
     {testfile},
@@ -67,7 +68,8 @@ const struct testlist {
     {testtime},
     {testud},
     {testuser},
-    {testvsn}
+    {testvsn},
+    {testskiplist}
 };
 
 #endif /* APR_TEST_INCLUDES */
