@@ -34,8 +34,8 @@ static void test_file_dup(abts_case *tc, void *data)
 
     /* First, create a new file, empty... */
     rv = apr_file_open(&file1, FILEPATH "testdup.file", 
-                       APR_READ | APR_WRITE | APR_CREATE |
-                       APR_DELONCLOSE, APR_OS_DEFAULT, p);
+                       APR_FOPEN_READ | APR_FOPEN_WRITE | APR_FOPEN_CREATE |
+                       APR_FOPEN_DELONCLOSE, APR_OS_DEFAULT, p);
     ABTS_INT_EQUAL(tc, APR_SUCCESS, rv);
     ABTS_PTR_NOTNULL(tc, file1);
 
@@ -65,8 +65,8 @@ static void test_file_readwrite(abts_case *tc, void *data)
 
     /* First, create a new file, empty... */
     rv = apr_file_open(&file1, FILEPATH "testdup.readwrite.file", 
-                       APR_READ | APR_WRITE | APR_CREATE |
-                       APR_DELONCLOSE, APR_OS_DEFAULT, p);
+                       APR_FOPEN_READ | APR_FOPEN_WRITE | APR_FOPEN_CREATE |
+                       APR_FOPEN_DELONCLOSE, APR_OS_DEFAULT, p);
     ABTS_INT_EQUAL(tc, APR_SUCCESS, rv);
     ABTS_PTR_NOTNULL(tc, file1);
 
@@ -76,7 +76,7 @@ static void test_file_readwrite(abts_case *tc, void *data)
 
     rv = apr_file_write(file3, TEST, &txtlen);
     ABTS_INT_EQUAL(tc, APR_SUCCESS, rv);
-    ABTS_INT_EQUAL(tc, sizeof(TEST), txtlen);
+    ABTS_SIZE_EQUAL(tc, sizeof(TEST), txtlen);
 
     fpos = 0;
     rv = apr_file_seek(file1, APR_SET, &fpos);
@@ -106,8 +106,8 @@ static void test_dup2(abts_case *tc, void *data)
     apr_status_t rv;
 
     rv = apr_file_open(&testfile, FILEPATH "testdup2.file", 
-                       APR_READ | APR_WRITE | APR_CREATE |
-                       APR_DELONCLOSE, APR_OS_DEFAULT, p);
+                       APR_FOPEN_READ | APR_FOPEN_WRITE | APR_FOPEN_CREATE |
+                       APR_FOPEN_DELONCLOSE, APR_OS_DEFAULT, p);
     ABTS_INT_EQUAL(tc, APR_SUCCESS, rv);
     ABTS_PTR_NOTNULL(tc, testfile);
 
@@ -128,6 +128,8 @@ static void test_dup2(abts_case *tc, void *data)
     rv = apr_file_dup2(errfile, saveerr, p);
     ABTS_INT_EQUAL(tc, APR_SUCCESS, rv);
     ABTS_PTR_NOTNULL(tc, errfile);
+
+    apr_file_close(saveerr);
 }
 
 static void test_dup2_readwrite(abts_case *tc, void *data)
@@ -141,8 +143,8 @@ static void test_dup2_readwrite(abts_case *tc, void *data)
     apr_off_t fpos;
 
     rv = apr_file_open(&testfile, FILEPATH "testdup2.readwrite.file", 
-                       APR_READ | APR_WRITE | APR_CREATE |
-                       APR_DELONCLOSE, APR_OS_DEFAULT, p);
+                       APR_FOPEN_READ | APR_FOPEN_WRITE | APR_FOPEN_CREATE |
+                       APR_FOPEN_DELONCLOSE, APR_OS_DEFAULT, p);
     ABTS_INT_EQUAL(tc, APR_SUCCESS, rv);
     ABTS_PTR_NOTNULL(tc, testfile);
 
@@ -161,7 +163,7 @@ static void test_dup2_readwrite(abts_case *tc, void *data)
     txtlen = sizeof(TEST2);
     rv = apr_file_write(errfile, TEST2, &txtlen);
     ABTS_INT_EQUAL(tc, APR_SUCCESS, rv);
-    ABTS_INT_EQUAL(tc, sizeof(TEST2), txtlen);
+    ABTS_SIZE_EQUAL(tc, sizeof(TEST2), txtlen);
 
     fpos = 0;
     rv = apr_file_seek(testfile, APR_SET, &fpos);
@@ -178,6 +180,8 @@ static void test_dup2_readwrite(abts_case *tc, void *data)
     rv = apr_file_dup2(errfile, saveerr, p);
     ABTS_INT_EQUAL(tc, APR_SUCCESS, rv);
     ABTS_PTR_NOTNULL(tc, errfile);
+
+    apr_file_close(saveerr);
 }
 
 abts_suite *testdup(abts_suite *suite)
